#include "config.h"
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/extensions/shape.h>
/* include Imlib2 stuff */
#include <Imlib2.h>
#include <ctype.h>
#include <errno.h>
#ifdef ENABLE_POWER_QUERY
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#endif
#ifdef ENABLE_HDDTEMP_QUERY
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#define GLOBALS_HERE
#include "global.h"
#include "procstat.h"
#include "devnames.h"


typedef struct {
  int nrow, ncol;
  int w;
  unsigned char **pre_cnt;
  char **intensity; /* > 0 for swap in, < 0 for swap out */
} SwapMatrix;

typedef struct IO_op_lst {
  enum { OP_READ, OP_WRITE } op;
  int n; /* "magnitude" of operation (log2 of kb/s) */
  int i,j; /* location of the IOMatrix */
  struct IO_op_lst *next;
} IO_op_lst;

#define CMAPSZ 256
typedef struct colormap {
  DATA32 p[CMAPSZ];
} cmap;

typedef struct {
  int w,h;
  float * __restrict * __restrict v; /* dim w+2, h+4 */
  cmap cm;
  IO_op_lst *ops;
} IOMatrix;

typedef struct {
  Display *display;
  Window   win, iconwin, rootwin;
  Visual *visual;
  Colormap colormap;
  int depth, screennum;
  Pixmap mask;		/* dockapp mask */
  Imlib_Image img;
  Imlib_Font bigfont,smallfont;
  int view_width, view_height, view_xpos, view_ypos;
  int win_width, win_height;

  unsigned update_display_delay_ms;
  unsigned update_stats_mult;
  
  unsigned char swap_matrix_luminosity, swap_matrix_lighting;

  SwapMatrix sm;
  IOMatrix iom;

  int nb_hd;
  enum {HD_ACTIVE, HD_STANDBY, HD_SLEEP, HD_UNKNOWN} *disk_power_mode;
  int *disk_temperature;

} Dock;

uid_t euid, uid;



void
update_swap_matrix(Dock *dock)
{
  SwapMatrix *sm = &dock->sm;
  unsigned i;
  int col,row;
  unsigned sw_in = ceil(get_swapin_throughput()*4), sw_out = ceil(get_swapout_throughput()*4);
  for (i=0; i < sw_in+sw_out; i++) {
    col = rand() % sm->ncol;
    row = rand() % sm->nrow;
    if (sm->intensity[row][col] == 0) {
      sm->pre_cnt[row][col] = rand() % 10; //(dock->update_stats_mult);
    }
    sm->intensity[row][col] = dock->swap_matrix_lighting * ((i < sw_in) ? -1 : +1);
  }
}

void update_io_matrix_rw(Dock *dock, float mbs, int op) {
  IOMatrix *io = &dock->iom;
  IO_op_lst *l;
  while (mbs>1e-5) {
    float v = MIN(mbs, 1024); /* a throughput > 2MB/s will create more than one spot */
    mbs -= v;
    l = calloc(1,sizeof(*l)); assert(l);
    l->next = io->ops;
    l->n = (int)(log2f(1+v*1024)/10.);
    l->op = op;
    l->i = rand() % io->h; 
    l->j = rand() % io->w;
    io->ops = l;
  }
}

void update_io_matrix(Dock *dock) {
  update_io_matrix_rw(dock, get_read_throughput(), OP_READ);
  update_io_matrix_rw(dock, get_write_throughput(), OP_WRITE);
}

void evolve_io_matrix(Dock *dock) {
  IOMatrix *io = &dock->iom;
  int i,j;
  float * __restrict pl, * __restrict tmp;
  IO_op_lst *o = io->ops, *po = NULL, *no;
  while (o) {
    io->v[o->i+1][o->j+1] = ((o->op == OP_READ) ? +1 : -1);
    no = o->next;
    if (--o->n <= 0) {
      if (po) { po->next = no; }
      else { io->ops = no; }
      free(o);
    }
    o = no;
  }
  
  /* brain-dead diffusion */
  pl = io->v[io->h+2];
  float *l = io->v[io->h+3];
  for (j=1; j < io->w+1; ++j) pl[j] = 0;  
  for (i=1; i < io->h+1; ++i) {
    for (j=1; j < io->w+1; ++j) {
      l[j] = io->v[i][j]*0.99 + (io->v[i][j-1] + io->v[i][j+1] + pl[j] + io->v[i+1][j] - 4*io->v[i][j])/5.;
    }
    tmp = pl; pl = io->v[i]; io->v[i] = l; l = tmp;
    io->v[io->h+2] = pl; io->v[io->h+3] = l;
  }
}

#define MERGEPIX(b,i,j,c,v) \
  { unsigned char *p=(unsigned char*)&(b)[((j)+(i)*dock->view_width)]+(c); \
    *p = v + ((255-v)* (*p))/255;                             \
  }

void
draw_swap_matrix(Dock *dock) 
{
  SwapMatrix *sm = &dock->sm;
  int rcnt[sm->nrow+1], ccnt[sm->ncol+1];
  int row, col;
  int isswapping = 0;
  static int darkcnt = 0;
  memset(rcnt, 0, sizeof rcnt); memset(ccnt, 0, sizeof ccnt);
  for (row = 0; row < sm->nrow; row++) {
    for (col = 0; col < sm->ncol; col++) {
      if (sm->pre_cnt[row][col]>0) { isswapping = 1; sm->pre_cnt[row][col]--; }
    }
  }
  if (isswapping) darkcnt = MIN(darkcnt+1,7);
  else darkcnt = MAX(darkcnt-1, 0);

  /* darken everything */
  if (darkcnt) {
    imlib_context_set_color(0, 0, 0, (darkcnt+1)*16);
    imlib_image_fill_rectangle(0, 0, dock->view_width, dock->view_height);
  }
  /* draw squares */
  for (row = 0; row < sm->nrow; row++) {
    for (col = 0; col < sm->ncol; col++) {
      if (sm->intensity[row][col] && sm->pre_cnt[row][col]==0) {
	int v = sm->intensity[row][col];
        v = (v * dock->swap_matrix_luminosity)/(int)dock->swap_matrix_lighting;
	rcnt[row  ] = MAX(rcnt[row  ],abs(v)); ccnt[col  ] = MAX(ccnt[col  ],abs(v));
	rcnt[row+1] = MAX(rcnt[row+1],abs(v)); ccnt[col+1] = MAX(ccnt[col+1],abs(v));
        //v = (v*155)/255;
        /*if (v > 0) imlib_context_set_color(100+v, 155-v, 155-v, 250);
          else     imlib_context_set_color(0, 100-v, 0, 250);*/
        if (v > 0) imlib_context_set_color(255, 50, 50, MIN(v+80,255));
        else     imlib_context_set_color(50, 255, 50, MIN(v+80,255));
        imlib_image_fill_rectangle(row*sm->w+1,col*sm->w+1,sm->w-1,sm->w-1);
	
	if (sm->intensity[row][col] > 0) {
	  sm->intensity[row][col]--;
	} else if (sm->intensity[row][col] < 0) {
	  sm->intensity[row][col]++;
	}
      }
    }
  }

  /* draw lines */
  for (row = 0; row < sm->nrow+1; row++) {
    if (rcnt[row]) {
      imlib_context_set_color(255, 255, 255, MIN(rcnt[row]*2/3, 255));
      imlib_image_draw_line(row*sm->w,0,row*sm->w,dock->view_width,0);
    }
  }
  for (col = 0; col < sm->ncol+1; col++) {
    if (ccnt[col]) {
      imlib_context_set_color(255, 255, 255, MIN(ccnt[col]*2/3, 255));
      imlib_image_draw_line(0,col*sm->w,dock->view_height,col*sm->w,0);
    }
  }
}


static void draw_io_matrix(Dock *dock, DATA32 * __restrict buff) {
  IOMatrix *io = &dock->iom;
  int i,j;
  for (i=0; i < io->h; ++i) {
    for (j=0; j < io->w; ++j) {  
      float vv = io->v[i+1][j+1];
      int v = ((int)ldexpf(vv,15)); /* (int)v*(2^15) */
      //if (v == 0) continue; 
      //if (v < 0) v*=3; /* write op are rare, so they are brighter .. */
      if (v > CMAPSZ/4) { /* cheaper than a log(vv) ... */
        v = MIN(CMAPSZ/4  + (v-CMAPSZ/4)/16,CMAPSZ/2-1);
      } else if (v < -CMAPSZ/4) {
        v = MAX(-CMAPSZ/4 + (v+CMAPSZ/4)/16,-CMAPSZ/2);
      }
      //assert(v+CMAPSZ/2>=0);
      //assert(v+CMAPSZ/2 < sizeof io->cm.p);
      buff[i+j*io->w] = io->cm.p[v+CMAPSZ/2];
    }
  }
}

static void query_hddtemp(Dock *dock) {
  int fd;
  struct hostent *h;
  struct sockaddr_in addr;
  char buff[1024];
  SET_VEC(dock->disk_temperature, -1, 0, dock->nb_hd);
  if ((h = gethostbyname("127.0.0.1")) == NULL) {
    fprintf(stderr, "gethostbyname(localhost) failed : %s\n", strerror(errno)); return;
  }
  if ((fd = socket(h->h_addrtype, SOCK_STREAM,0)) == -1) {
    fprintf(stderr, "can't create socket : %s\n", strerror(errno)); return;
  }
  memset(&addr,0,sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(Prefs.hddtemp_port);
  addr.sin_addr.s_addr = ((struct in_addr*)(h->h_addr))->s_addr;
  if (connect(fd, (struct sockaddr *)&addr, sizeof addr) < 0) {
    close(fd);
    fprintf(stderr,"can't connect to 127.0.0.1:%d : %s\n", Prefs.hddtemp_port, strerror(errno)); return;
  }
  int n = 0;
  do {
    int nn=read(fd,buff+n,MAX((int)(sizeof buff) - n,0));
    if (nn <= 0) { if (nn < 0) n=nn; break; }
    n+=nn;
  } while (n != sizeof buff);

  BLAHBLAH(2,printf("n=%d, err=%s\n",n, strerror(errno)));
  if (n != -1) {
    char *s;
    DiskList *dl;
    int cnt;
    buff[MIN(n,(int)sizeof buff - 1)] = 0;
    for (dl = first_hd_in_list(), cnt = 0; dl; dl = next_hd_in_list(dl), cnt++) {
      if (dl->enable_hddtemp) {
        int found = 0;
        if ((s=strstr(buff, dl->name))) { found = 1; s+= strlen(dl->name); }
        else if ((s=strstr(buff, dl->devfsname))) { found = 1; s+= strlen(dl->devfsname); }
        if (found) {
          s=strchr(s,'|');
          if (s && ((s=strchr(s+1,'|')))) {
            char *p = strchr(++s,'|'), oldv=0; if (p) { oldv = *p; *p = 0; }
            BLAHBLAH(1,printf("temperature of '%s' : %s\n", dl->name, s));
            if (strcmp(s,"SLP") == 0) dock->disk_temperature[cnt] = -2;
            else dock->disk_temperature[cnt] = atoi(s);
            if (p) *p = oldv; /* or bug... */
          }
        } else {
          ONLY_NTIMES(4,fprintf(stderr, "could not find device '%s' in the output of hddtemp: '%s'\n", dl->name, buff));
        }
      }
    }
  } else { fprintf(stderr, "error with hddtemp: %s\n", strerror(errno)); }  
  close(fd);
}

const char *power_mode_str(int m) {
  switch (m) {
  case HD_ACTIVE: return "active/idle";
  case HD_STANDBY: return "standby";
  case HD_SLEEP: return "sleep";
  case HD_UNKNOWN: 
  default:
    return "unknown/error";
  }
}

/* stolen from hdparm */
#ifdef ENABLE_POWER_QUERY
int query_power_mode(const char *devname) {
  int state;
  int fd;
  if (seteuid(euid) == -1) { perror("seteuid: Cannot gain root privileges"); return HD_UNKNOWN; }
  fd = open(devname, O_RDONLY|O_NONBLOCK); /* won't work if not root .. */
  if (fd == -1) {
    fprintf(stderr, "can't get %s power status: %s\n", devname, strerror(errno));
    return HD_UNKNOWN;
  } else {
    

#ifndef WIN_CHECKPOWERMODE1
#define WIN_CHECKPOWERMODE1 0xE5
#endif
#ifndef WIN_CHECKPOWERMODE2
#define WIN_CHECKPOWERMODE2 0x98
#endif
    unsigned char args[4] = {WIN_CHECKPOWERMODE1,0,0,0};
    //no_scsi(); /* this is checked before.. */
    if (ioctl(fd, HDIO_DRIVE_CMD, &args)
        && (args[0] = WIN_CHECKPOWERMODE2) /* try again with 0x98 */
        && ioctl(fd, HDIO_DRIVE_CMD, &args)) {
      if (errno != EIO || args[0] != 0 || args[1] != 0)
        state = HD_UNKNOWN;
      else
        state = HD_SLEEP;
    } else {
      state = (args[2] == 255) ? HD_ACTIVE : HD_STANDBY;
    }
  }
  close(fd);
  seteuid(uid);
  BLAHBLAH(1,printf("%s state is:  %s\n", devname, power_mode_str(state)));
  return state;
}
#endif

static void draw(Dock *dock)
{
  int w,h;
  char s[200];
  DATA32 *buff = imlib_image_get_data();
  static int tpstep = 0;
  char tpmsg[20];

  memset(buff, 0, 4*dock->view_width*dock->view_height);
  if (!Prefs.disable_io_matrix) {
    evolve_io_matrix(dock);
    draw_io_matrix(dock,buff);
  }
  imlib_image_put_back_data(buff);

  if (!Prefs.disable_hd_list) {
    DiskList *dl;
    int y = dock->view_height, cnt;
    int wnames = 0;
    imlib_context_set_font(dock->smallfont);
    for (dl = first_hd_in_list(), cnt=0; dl; dl = dl->next, cnt++) {
      strncpy(s, dl->name, sizeof s); s[5] = 0;
      imlib_get_text_size(s,&w,&h);
      wnames = MAX(wnames,w+5);
    }
    if (Prefs.enable_hddtemp) wnames = dock->view_width;
    for (dl = first_hd_in_list(), cnt=0; dl; dl = dl->next, cnt++) {
      strncpy(s, dl->name, sizeof s); s[5] = 0;
      imlib_get_text_size(s,&w,&h);
      imlib_context_set_color(100, 100, 100, 150);
      if (Prefs.enable_hddtemp) {
        imlib_image_fill_rectangle(0,y-h+1,wnames,h);
      } else {
        imlib_image_fill_rectangle(0,y-h+1,wnames+1,h);
        imlib_context_set_color(100, 100, 100, 200); 
        imlib_image_draw_line(wnames+1,y-h,wnames+1,y-1,0);
      }
      if (!Prefs.disable_hd_leds) {
        if (dl->touched_r) { imlib_context_set_color(50, 255, 50, 25*dl->touched_r--); imlib_image_fill_rectangle(1,y-5,3,3); }
        if (dl->touched_w) { imlib_context_set_color(255,100-10*dl->touched_w,100-10*dl->touched_w, 25*dl->touched_w--); imlib_image_fill_rectangle(1,y-9,3,3); }
      }
      y -= h;
      imlib_context_set_color(200, 255, 255, 200);
      imlib_text_draw(5,y,s);
#ifdef ENABLE_POWER_QUERY
      if (dock->disk_power_mode[cnt] == HD_SLEEP ||
          dock->disk_power_mode[cnt] == HD_STANDBY) {
        if (dock->disk_power_mode[cnt] == HD_SLEEP) strcpy(s,"sleep"); else strcpy(s,"stdby");
        imlib_get_text_size(s,&w,&h);
        imlib_context_set_color(100, 255, 255, 255);
        imlib_text_draw(dock->view_width - w-1,y, s);
      } else 
#endif
      if (dock->disk_temperature[cnt] != -1) {
        if (dock->disk_temperature[cnt] != -2) snprintf(s,200,"%d",dock->disk_temperature[cnt]);
        else strcpy(s,"SLP");
        imlib_get_text_size(s,&w,&h);
        imlib_context_set_color(255, 250, 100, 255);
        imlib_text_draw(dock->view_width - w - 6,y, s);
        imlib_context_set_color(255, 255, 255, 200);
        imlib_image_draw_ellipse(dock->view_width - 3, y + h/4, 1, 1);
      }
    }
    imlib_context_set_color(100, 100, 100, 200); 
    imlib_image_draw_line(0,y,wnames,y,0);
  }
  if (!Prefs.disable_swap_matrix) draw_swap_matrix(dock);

  if (get_read_mean_throughput() + get_write_mean_throughput() > Prefs.popup_throughput_threshold) {
    tpstep = MIN(tpstep+1,4);
    snprintf(tpmsg,sizeof tpmsg, "%.1fM/s",get_read_mean_throughput() + get_write_mean_throughput());
  } else if (tpstep) tpstep--;

  if (tpstep) {
    int ww;
    imlib_context_set_font(dock->bigfont);
    imlib_get_text_size(tpmsg,&w,&h); ww = dock->view_width+2; //((w + 3 + 4)/5)*5;
    imlib_context_set_color(128, 128, 128, tpstep*30);
    imlib_image_draw_rectangle(-1,-1,ww,h+1);
    imlib_context_set_color(128, 128, 128, 10+tpstep*25);
    imlib_image_fill_rectangle(0,0,ww-1,h);
    imlib_context_set_color(255, 255, 255, 50+tpstep*50);
    imlib_text_draw((dock->view_width - w)/2, 0, tpmsg);
  }

  /*  if (strcmp(dock->disk_power_mode,"active"))
      imlib_text_draw(1,dock->view_height/2, dock->disk_power_mode);
  if (dock->disk_temperature != -1) {
    if (dock->disk_temperature != -2) snprintf(s,200,"%d",dock->disk_temperature);
    else strcpy(s,"SLP");
    imlib_get_text_size(s,&w,&h);
    imlib_context_set_color(255, 250, 100, 255);
    imlib_text_draw(dock->view_width - w -6,dock->view_height-h, s);
    imlib_context_set_color(255, 255, 255, 200);
    if (dock->disk_temperature != -2) imlib_image_draw_ellipse(dock->view_width - 3, dock->view_width - h + 4, 1, 1);
  }
  */
}


static void event_loop(Dock *dock)
{
  int tic_cnt = 0;
  while (1) {
    tic_cnt++;
    /*    do {
      XNextEvent(disp, &ev);
      } while (XPending(dock->display));*/
    if (tic_cnt % dock->update_stats_mult == 0) {
      update_stats();
      if (!Prefs.disable_io_matrix)   update_io_matrix(dock);
      if (!Prefs.disable_swap_matrix) update_swap_matrix(dock);
    }
    if (tic_cnt % 100 == 5) {
#ifdef ENABLE_HDDTEMP_QUERY
      if (Prefs.enable_hddtemp) {
        query_hddtemp(dock);
      }
#endif
#ifdef ENABLE_POWER_QUERY
      if (Prefs.enable_power_status) {
        DiskList *dl; int cnt;
        for (cnt = 0, dl = first_hd_in_list(); dl; dl = next_hd_in_list(dl), ++cnt) {
          if (dl->enable_power_status && !dl->is_scsi) {
            char devname[512]; snprintf(devname, 512, "/dev/%s", dl->name);
            dock->disk_power_mode[cnt] = query_power_mode(devname);
          }
        }
      }
#endif
    }
    //if (tic_cnt > 1500) exit(1);
    usleep(dock->update_display_delay_ms * 1000);
    draw(dock);
    imlib_render_image_on_drawable(dock->view_xpos, dock->view_ypos);
  }
}

static Imlib_Font *load_font(char *prefname, char **flist) {
  Imlib_Font *f = NULL;
  
  if (prefname) {
    f = imlib_load_font(prefname);
    if (!f) {
      int i,np; char **s;
      fprintf(stderr, "warning: could not find font '%s' in the font path:\n",prefname);
      s = imlib_list_font_path(&np);
      for (i=0; i < np; ++i) fprintf(stderr, "  %s\n", s[i]);
    } else {
      BLAHBLAH(1,printf("loaded font %s\n", *flist));
    }
  }
  if (!f) {
    for (; *flist; ++flist) {
      if ((f = imlib_load_font(*flist))) {
        BLAHBLAH(1,printf("loaded font %s\n", *flist)); break;
      }
    }
    if (!f) {
      fprintf(stderr, "could not load a small ttf font .. use the -f option\n"); 
      exit(1); 
    }
  }
  return f;
}

static void dock_xinit(Dock *dock, int argc, char **argv) {
  XSizeHints xsh;
  XWMHints *xwmh;
  XClassHint xch;
  XTextProperty xtp;
  int i;
  dock->display = XOpenDisplay(getenv("DISPLAY"));
  if(!dock->display) { fprintf(stderr, "Couldn't connect to display\n"); exit(1); }
  dock->visual = DefaultVisual(dock->display, DefaultScreen(dock->display));
  dock->depth = DefaultDepth(dock->display, DefaultScreen(dock->display));
  dock->colormap = DefaultColormap(dock->display, DefaultScreen(dock->display));
  dock->screennum = DefaultScreen(dock->display);
  dock->rootwin = RootWindow(dock->display, dock->screennum);

  /* set size hints 64 x 64 */
  xsh.flags = USSize | USPosition;
  xsh.width = 64;
  xsh.height = 64;
  XWMGeometry(dock->display, dock->screennum, "64x64+0+0", NULL, 0,
              &xsh, &xsh.x, &xsh.y, &xsh.width, &xsh.height, &i);

  /* create the application window */
  dock->win = XCreateSimpleWindow(dock->display, dock->rootwin,
				  xsh.x, xsh.y, xsh.width, xsh.height, 0,
				  BlackPixel(dock->display, dock->screennum),
				  WhitePixel(dock->display, dock->screennum));
  
  if(!dock->win) {
    fprintf(stderr, "Couldn't create window\n");
    exit(1);
  }
  /* create icon window */
  dock->iconwin = XCreateSimpleWindow(dock->display, dock->rootwin,
                                      xsh.x, xsh.y, xsh.width, xsh.height, 0,
                                      BlackPixel(dock->display, dock->screennum),
                                      WhitePixel(dock->display, dock->screennum));
  if(!dock->iconwin) {
    fprintf(stderr, "Couldn't create icon window\n");
    exit(1);
  }
    
  {
    char *mask; 
    ALLOC_VEC(mask,dock->win_height*dock->win_width/8);
    dock->mask = XCreateBitmapFromData(dock->display, dock->win, mask, dock->win_width, dock->win_height);
    GC gc = XCreateGC(dock->display, dock->mask, 0, NULL);
    XSetForeground(dock->display, gc, WhitePixel(dock->display, dock->screennum));
    XFillRectangle(dock->display, dock->mask, gc, dock->view_xpos, dock->view_ypos, dock->view_width, dock->view_height);
    XFreeGC(dock->display,gc);
  }
  /* setup shaped window */
  XShapeCombineMask(dock->display, dock->win, ShapeBounding, 
                    0, 0, dock->mask, ShapeSet);
  XShapeCombineMask(dock->display, dock->iconwin, ShapeBounding,
                    0, 0, dock->mask, ShapeSet);
  /* set window manager hints */
  xwmh = XAllocWMHints();
  xwmh->flags = WindowGroupHint | IconWindowHint | StateHint;
  xwmh->icon_window = dock->iconwin;
  xwmh->window_group = dock->win;
  xwmh->initial_state = WithdrawnState;
  XSetWMHints(dock->display, dock->win, xwmh);
  XFree(xwmh); xwmh = NULL;
  /* set class hints */
  xch.res_name = argv[0];
  xch.res_class = argv[0];
  XSetClassHint(dock->display, dock->win, &xch);  
  /* set size hints */
  XSetWMNormalHints(dock->display, dock->win, &xsh);
  /* tell window manager app name */
  if(!XStringListToTextProperty(&argv[0], 1, &xtp)) {
    fprintf(stderr, "Couldn't create text property\n");
    exit(1);
  }
  XSetWMName(dock->display, dock->win, &xtp);
  XFree(xtp.value); 
  /* set the command line for restarting */
  XSetCommand(dock->display, dock->win, argv, argc);  
  /* map the main window */
  XMapWindow(dock->display, dock->win);

   /* set the maximum number of colors to allocate for 8bpp and less to 32 */
  imlib_set_color_usage(32);
  /* dither for depths < 24bpp */
  imlib_context_set_dither(1);
  /* set the display , visual, colormap and drawable we are using */
  imlib_context_set_display(dock->display);
  imlib_context_set_visual(dock->visual);
  imlib_context_set_colormap(dock->colormap);
  imlib_context_set_drawable(dock->iconwin);
  dock->img = imlib_create_image(dock->view_width,dock->view_height);
  imlib_context_set_image(dock->img);
  {
    char fp[512];
    char *bigfontlist[] = {"Arial_Black/10", "FreeSansBold/11", "VeraMoBd/9", NULL};
    char *smallfontlist[] = {"Vera/7","FreeSans/7","Trebuchet_MS/7",NULL};
    snprintf(fp, 512, "%s/.fonts", getenv("HOME"));
    imlib_add_path_to_font_path(fp);
    imlib_add_path_to_font_path("/usr/share/fonts/truetype");
    imlib_add_path_to_font_path("/usr/share/fonts/truetype/freefont");
    imlib_add_path_to_font_path("/usr/share/fonts/truetype/ttf-bitstream-vera");
    dock->bigfont = load_font(Prefs.bigfontname, bigfontlist);
    dock->smallfont = load_font(Prefs.smallfontname, smallfontlist);
  }
  imlib_context_set_TTF_encoding(IMLIB_TTF_ENCODING_ISO_8859_1);
  //imlib_set_cache_size(0);imlib_set_font_cache_size(0);
}

void setup_cmap(cmap *m) {
  struct {
    float x0; 
    DATA32 c;
  } colors0[] = {{-128, 0xff8080},
                 {- 80, 0xff0000},
                 { -40, 0xff0080},
                 { -20, 0x800000},
                 {   0, 0x000000},
                 {  10, 0x008000},
                 {  60, 0xf09000},
                 {  90, 0xffa000},
                 { 116, 0xffd000},
                 { 127, 0xffff00}},
    colors1[] = {{-128, 0xff0000},
                 {- 64, 0x808080},
                 { 0, 0x404040},
                 //{  , 0x000000},
                 //{  0 , 0x000000},
                 {  1, 0x208020},
                 {  64, 0x509050},
                 {+ 90, 0x60C060},
                 {+127, 0x008000}},
      colors2[] = {{-128, 0x400000},
                   { -60, 0xA00000},
                   { -34, 0xff0000},
                   { -16, 0x400000},
                   {   0, 0x000000},
                   {  16, 0x000040},
                   {  34, 0x0000ff},
                   {  60, 0x0000A0},
                   {+128, 0x000040}}, 
      colors3[] = {{-128, 0x500060},
                   { -60, 0x500050},
                   { -34, 0x000000},
                   {   0, 0x000000},
                   {  34, 0x000000},
                   {  60, 0x206020},
                   {+128, 0x205020}},

    *cdef = NULL;

#define SELMAP(n) \
if (Prefs.iomatrix_colormap == n) { sz = sizeof(colors##n)/sizeof(*colors0); cdef = colors##n; }

  unsigned i, sz=0;
  SELMAP(0); SELMAP(1); SELMAP(2); SELMAP(3);
  float x0 = cdef[0].x0, x1 = cdef[sz-1].x0;
  for (i = 0; i < sz - 1; ++i) {
    int i0 = (int)((cdef[i].x0-x0)*CMAPSZ/(x1-x0));
    int i1 = (int)((cdef[i+1].x0-x0)*CMAPSZ/(x1-x0));
    int r1 = cdef[i].c>>16 & 0xff, r2 = cdef[i+1].c>>16 & 0xff;
    int g1 = cdef[i].c>> 8 & 0xff, g2 = cdef[i+1].c>> 8 & 0xff;
    int b1 = cdef[i].c     & 0xff, b2 = cdef[i+1].c     & 0xff;
    int j;
    for (j=i0; j <= MIN(i1,CMAPSZ-1); ++j) {
      float alpha = (j-i0+.5)/(float)(i1-i0);
      m->p[j] = 
        (MIN((int)(r1*(1-alpha) + alpha*r2),255)<<16) +
        (MIN((int)(g1*(1-alpha) + alpha*g2),255)<<8) +
        (MIN((int)(b1*(1-alpha) + alpha*b2),255));
      //printf("cmap[%d] = 0x%06x\n", j, m->p[j]);
    }
  }
}

void parse_options(int argc, char **argv) {
  int d_opt_used = 0;
  const char *help = 
    "wmhd 1.0 - monitor harddrive (or partition) activity\n"
    "A recent kernel is required (>2.4.20)\n"
    "Option list:\n"
    "  -h : print this.\n"
    "  -v : increase verbosity\n"
    "  -V : print version\n"
    "  -t[port] : display hd temperatures (requires hddtemp daemon running)\n"
    /*    "  -p : monitor harddrive power status (idle, standby or sleeping, as reported by hdparm -C)\n"
          "       note that this options requires that wmhd is suid..\n"*/ /* hddtemp already reports when a drive is asleep ... */
    "  -i : disable the animation reflecting swap activity\n"
    "  -s : disable background animation reflecting disk activity\n"
    "  -l : hide the small led indicating disk activity\n"
    "  -F path : add a new directory to the font search directory list\n"
    "  -f fontname/size\n"
    "  -g fontname/size : set the 'small font' and the 'big font' names\n"
    "       (default: -f Vera/7 -g Arial_Black/10)\n"
    "  -L : hide completely the list of drives (with leds and temperatures..)\n"
    "  -c n : select colormap number n\n" 
    "  -P mbs : minimum io throughput (MB/s) that will be displayed in the popup display (default 0.5)\n"
    "  -d device : monitor activity on the specified hard-drive. This option can be\n"
    "       used many times. If you do not use it, all hard-drives will be monitored.\n"
    "       The device may also be a single partition.\n";
  Prefs.verbosity = 0;
  Prefs.enable_hddtemp = 0;
  Prefs.hddtemp_port = 7634;
  Prefs.enable_power_status = 0;
  Prefs.disable_swap_matrix = 0;
  Prefs.disable_io_matrix = 0;
  Prefs.disable_hd_leds = 0;
  Prefs.disable_hd_list = 0;
  Prefs.iomatrix_colormap = 0;
  Prefs.debug_swapio = 0; Prefs.debug_disk_wr = Prefs.debug_disk_rd = 0;
  Prefs.popup_throughput_threshold = 0.5; /* MB/s */
  Prefs.bigfontname = Prefs.smallfontname = NULL;
  while (1) {
    int c;
    c = getopt(argc, argv, "hvVt::pislF:f:g:Lc:P:d:@:");
    if (c == -1)
      break;
    switch (c) {
    case ':':
    case '?': 
      exit(1);
    case 'h':
      puts(help);
      exit(0);
    case 'v':
      Prefs.verbosity++;
      break;
    case 'V':
      printf("wmhdplop %s\n",VERSION);
      break;
    case 't':
      Prefs.enable_hddtemp = 1;
      if (optarg) Prefs.hddtemp_port = atoi(optarg);
      break;
    case 'p':
      Prefs.enable_power_status = 1;
      break;
    case 'i':
      Prefs.disable_io_matrix = 1;
      break;
    case 's':
      Prefs.disable_swap_matrix = 1;
      break;
    case 'l':
      Prefs.disable_hd_leds = 1;
      break;
    case 'F':
      imlib_add_path_to_font_path(optarg);
      break;
    case 'f':
      Prefs.smallfontname = strdup(optarg);
      break;
    case 'g':
      Prefs.bigfontname = strdup(optarg);
      break;
    case 'L':
      Prefs.disable_hd_list = 1;
      break;
    case 'c':
      Prefs.iomatrix_colormap = atoi(optarg);
      if (Prefs.iomatrix_colormap > 3) { fprintf(stderr,"invalid colormap number\n"); exit(1); }
      break;
    case 'P':
      Prefs.popup_throughput_threshold = atof(optarg);
      break;
    case 'd':
      assert(optarg);
      BLAHBLAH(1,printf("adding device %s to monitored disc list\n",optarg));
      add_device(optarg);
      d_opt_used = 1;
      break;
    case '@':
      sscanf(optarg,"%d,%d,%d",&Prefs.debug_swapio,&Prefs.debug_disk_rd,&Prefs.debug_disk_wr);
      break;
    default:
      assert(0);
    }
  }
  if (optind != argc) {
    fprintf(stderr, "unknown option: %s\n", argv[optind]); exit(1);
  }
  if (!d_opt_used) scan_all_hd();
}


int
main(int argc, char**argv)
{
  Dock *dock;
  euid = geteuid(); uid = getuid(); seteuid(uid);
  ALLOC_OBJ(dock);
  srand(time(NULL));
  
  parse_options(argc,argv);
  if (nb_dev_in_list() == 0) {
    fprintf(stderr, "please use option -d to had devices (i.e. wmhd -d /dev/hda -d /dev/hdb ...)\n");
    exit(1);
  }
  dock->view_xpos = 4;
  dock->view_ypos = 4;
  dock->view_width = 56;
  dock->view_height = 56;
  dock->win_width = 64;
  dock->win_height = 64;

  dock_xinit(dock,argc,argv);

  dock->update_display_delay_ms = 40;
  dock->update_stats_mult = 2;

  dock->swap_matrix_lighting = (int)(300/dock->update_display_delay_ms);
  dock->swap_matrix_luminosity = 255;

  dock->nb_hd = nb_hd_in_list();
  ALLOC_VEC(dock->disk_power_mode, dock->nb_hd); SET_VEC(dock->disk_power_mode, HD_ACTIVE,0,dock->nb_hd);
  ALLOC_VEC(dock->disk_temperature, dock->nb_hd); SET_VEC(dock->disk_temperature, -1, 0, dock->nb_hd);

  init_stats(dock->update_display_delay_ms*1e-3*dock->update_stats_mult);

  dock->sm.w = 6; 
  dock->sm.nrow = (dock->view_width-1) / dock->sm.w;
  dock->sm.ncol = (dock->view_height-1) / dock->sm.w;
  ALLOC_ARR(dock->sm.pre_cnt, dock->sm.nrow, dock->sm.ncol);
  ALLOC_ARR(dock->sm.intensity, dock->sm.nrow, dock->sm.ncol);

  dock->iom.w = dock->view_width; dock->iom.h = dock->view_height;
  ALLOC_ARR(dock->iom.v,dock->iom.h+4, dock->iom.w+2);
  dock->iom.ops = NULL;
  setup_cmap(&dock->iom.cm);
  event_loop(dock);
  return 0;
}
